/*
* JBoss, Home of Professional Open Source.
* Copyright 2006, Red Hat Middleware LLC, and individual contributors
* as indicated by the @author tags. See the copyright.txt file in the
* distribution for a full listing of individual contributors. 
*
* This is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2.1 of
* the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free
* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
* 02110-1301 USA, or see the FSF site: http://www.fsf.org.
*/ 
package org.jboss.dependency.util;

import java.io.Serializable;
import java.util.Collection;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

/**
 * A set working concurrently like CopyOnWriteArraySet apart from that its elements are not sorted. It is
 * backed by a {@link FastCopyHashMap}.
 * 
 * @author <a href="kabir.khan@jboss.com">Kabir Khan</a>
 * @version $Revision: 1.1 $
 */
public class FastCopyOnWriteSet<E> implements Set<E>, Cloneable, Serializable
{
   private static final long serialVersionUID = 1L;

   private volatile FastCopyHashMap<E, Object> delegate;
   
   private static final Object VALUE = new Object();

   //TODO add other ctors from FCHM
   
   public FastCopyOnWriteSet()
   {
      delegate = new FastCopyHashMap<E, Object>();
   }

   /**
    * This method should NOT be used. It is there for testing only
    */
   public Map<E, Object> getDelegate()
   {
      return delegate;
   }

   public boolean add(E e)
   {
      boolean added;
      synchronized (this)
      {
         FastCopyHashMap<E, Object> newDelegate = new FastCopyHashMap<E, Object>(delegate);
         added = newDelegate.put(e, VALUE) == null;
         if (added)
            delegate = newDelegate;
      }
      return added;
   }

   public boolean addAll(Collection<? extends E> c)
   {
      boolean added = false;
      synchronized (this)
      {
         FastCopyHashMap<E, Object> newDelegate = new FastCopyHashMap<E, Object>(delegate);
         for (E e : c)
         {
            boolean didAdd = newDelegate.put(e, VALUE) == null;
            if (!added && didAdd)
               added = true;
         }
         if (added)
            delegate = newDelegate;
      }
      return added;
   }

   public void clear()
   {
      synchronized (this)
      {
         if (delegate.isEmpty() == false)
         {
            FastCopyHashMap<E, Object> newDelegate = new FastCopyHashMap<E, Object>();
            delegate = newDelegate;
         }
      }
   }

   public boolean remove(Object o)
   {
      boolean removed = false;
      synchronized (this)
      {
         FastCopyHashMap<E, Object> newDelegate = new FastCopyHashMap<E, Object>(delegate);
         removed = newDelegate.remove(o) != null;
         if (removed)
            delegate = newDelegate;
      }
      return removed;
   }

   public boolean removeAll(Collection<?> c)
   {
      boolean removed = false;
      synchronized (this)
      {
         FastCopyHashMap<E, Object> newDelegate = new FastCopyHashMap<E, Object>(delegate);
         for (Object o : c)
         {
            boolean didRemove = newDelegate.remove(o) != null;
            if (!removed && didRemove)
               removed = true;
         }
         if (removed)
            delegate = newDelegate;
      }
      return removed;
   }

   public boolean retainAll(Collection<?> c)
   {
      boolean changed = false;
      synchronized (this)
      {
         FastCopyHashMap<E, Object> newDelegate = new FastCopyHashMap<E, Object>();
         
         for (E e : delegate.keySet())
         {
            if (c.contains(e))
               newDelegate.put(e, VALUE);
         }
         changed = newDelegate.size() != delegate.size();
         if (changed)
            delegate = newDelegate;
      }
      return changed;
   }

   public boolean containsAll(Collection<?> c)
   {
      FastCopyHashMap<E, Object> old = delegate;
         for (Object o : c)
         {
            if (old.containsKey(o) == false)
               return false;
         }
      return true;
   }

   public boolean contains(Object o)
   {
      return delegate.containsKey(o);
   }

   public boolean isEmpty()
   {
      return delegate.isEmpty();
   }

   public Iterator<E> iterator()
   {
      return delegate.keySet().iterator();
   }

   public int size()
   {
      return delegate.size();
   }

   public Object[] toArray()
   {
      return delegate.keySet().toArray();
   }

   public <T> T[] toArray(T[] a)
   {
      return delegate.keySet().toArray(a);
   }

   @Override
   @SuppressWarnings("unchecked")
   public boolean equals(Object o)
   {
      if (o instanceof Collection == false)
         return false;
      return delegate.keySet().equals(o);
   }
}
